# lib: built-in
from typing import (
    List,
    AsyncGenerator,
    Set,
)
import re
from pathlib import Path

# lib: external
from watchfiles import (
    Change,
    DefaultFilter,
    awatch,
)
from watchfiles.main import FileChange
from aiostream import stream

# lib: local
from .dataclass import (
    BuildConfig,
    FileProcessInfo,
    FileChangeResult,
)

from . import process


class WatchFilter(DefaultFilter):
    def __init__(self,
            *args,
            dir_base: Path,
            list_regex: List[re.Pattern] = [],
            exclude_globs: List[str],
            **kw,
    ):
        self.dir_base = dir_base
        self.list_regex = list_regex
        self.exclude_globs = exclude_globs
        super().__init__(*args, **kw)

    def __call__(self, change: Change, path: str) -> bool:
        path_rel = Path(path).relative_to(self.dir_base)
        _is_valid_path = process.is_valid_path(
            path=path_rel,
            list_regex=self.list_regex,
            exclude_globs=self.exclude_globs,
        )

        return (
            super().__call__(change, path) and _is_valid_path
        )


async def handle_async_html_list_change(
        build_config: BuildConfig,
        async_html_list_file_change: AsyncGenerator[Set[FileChange]]
) -> AsyncGenerator[List[FileChangeResult]]:

    async_list_file_change = (
        list_file_change
        async for list_file_change
        in async_html_list_file_change
    )

    async for list_file_change in async_list_file_change:
        list_file_change_result: list[FileChangeResult] = []
        for change, path in list_file_change:
            type = 'html'
            if path.endswith('.html'):
                file_process_info = FileProcessInfo(
                    path=Path(path),
                    dir_src=Path(build_config.dir_src),
                    dir_dest=Path(build_config.dir_dest),
                )
                if change == Change.deleted:
                    process.delete_file(file_process_info)
                elif ((change == Change.modified)
                    or (change == Change.added)):
                    process.build_html(file_process_info)
            elif path.endswith('.md'):
                type = 'markdown'

            path_rel = Path(path).relative_to(
                Path(build_config.dir_src).resolve()
            )
            path_rel = Path(build_config.dir_src) / path_rel
            list_file_change_result.append(
                FileChangeResult(
                    path=str(path_rel),
                    type=type,
                    change=change,
                )
            )
        yield list_file_change_result

async def handle_async_copy_list_change(
        build_config: BuildConfig,
        async_copy_list_file_change: AsyncGenerator[Set[FileChange]]
) -> AsyncGenerator[List[FileChangeResult]]:

    async_list_file_change = (
        list_file_change
        async for list_file_change
        in async_copy_list_file_change
    )

    async for list_file_change in async_list_file_change:
        list_file_change_result: list[FileChangeResult] = []
        for change, path in list_file_change:
            file_process_info = FileProcessInfo(
                path=Path(path),
                dir_src=Path(build_config.dir_src),
                dir_dest=Path(build_config.dir_dest),
            )
            if change == Change.deleted:
                process.delete_file(file_process_info)
            elif ((change == Change.modified)
                or (change == Change.added)):
                process.copy_file(file_process_info)

            path_rel = Path(path).relative_to(
                Path(build_config.dir_src).resolve()
            )
            path_rel = build_config.dir_src / path_rel
            list_file_change_result.append(
                FileChangeResult(
                    path=str(path_rel),
                    type='copy',
                    change=change,
                )
            )
        yield list_file_change_result


async def handle_async_watch_list_change(
        build_config: BuildConfig,
        async_watch_list_file_change: AsyncGenerator[Set[FileChange]]
) -> AsyncGenerator[List[FileChangeResult]]:

    async_list_file_change = (
        list_file_change
        async for list_file_change
        in async_watch_list_file_change
    )

    async for list_file_change in async_list_file_change:
        list_file_change_result: list[FileChangeResult] = []
        for change, path in list_file_change:
            path_rel = Path(path).relative_to(
                Path(build_config.dir_dest).resolve()
            )
            path_rel = Path(build_config.dir_dest) / path_rel
            list_file_change_result.append(
                FileChangeResult(
                    path=str(path_rel),
                    type='watch',
                    change=change,
                )
            )
        yield list_file_change_result

async def run(build_config: BuildConfig) -> AsyncGenerator[List[FileChangeResult]]:
    list_html_regex = [re.compile(r'.*\.html$'), re.compile(r'.*\.md$')]
    list_copy_regex = [re.compile(copy_regex) for copy_regex in build_config.copy]
    list_watch_regex = [re.compile(watch_regex) for watch_regex in build_config.watch]

    async_html_list_change = awatch(
        build_config.dir_src,
        watch_filter=WatchFilter(
            dir_base=Path(build_config.dir_src).resolve(),
            list_regex=list_html_regex,
            exclude_globs=build_config.exclude,
        )
    )

    async_copy_list_change = awatch(
        build_config.dir_src,
        watch_filter=WatchFilter(
            dir_base=Path(build_config.dir_src).resolve(),
            list_regex=list_copy_regex,
            exclude_globs=build_config.exclude,
        )
    )

    async_watch_list_change = awatch(
        Path.cwd(),
        watch_filter=WatchFilter(
            dir_base=Path.cwd().resolve(),
            list_regex=list_watch_regex,
            exclude_globs=build_config.exclude,
        )
    )

    stream_watch = stream.merge(
        handle_async_html_list_change(
            build_config,
            async_html_list_change
        ),
        handle_async_copy_list_change(
            build_config,
            async_copy_list_change
        ),
        handle_async_watch_list_change(
            build_config,
            async_watch_list_change
        ),
    )

    async with stream_watch.stream() as streamer:
        async for watch_result in streamer:
            yield watch_result



# lib: built-in
from pathlib import Path
from glob import iglob
import re
from itertools import chain

# lib: external
from loguru import logger

# lib: local
from . import process
from .dataclass import BuildConfig, FileProcessInfo


def run(build_config: BuildConfig) -> None:
    """
    Build HTML files from Jinja2 templates.

    Args:
        build_info: Build information containing source and destination directories
    """

    dir_src = Path(build_config.dir_src)
    dir_dest = Path(build_config.dir_dest)

    # Create destination directory if it doesn't exist
    dir_dest.mkdir(parents=True, exist_ok=True)
    logger.info(f"🔍 Looking for files in: {dir_src}/")
    logger.info(f"📤 Output directory: {dir_dest}/")

    list_copy_regex = [re.compile(regex) for regex in build_config.copy]

    gen_html = filter(lambda path:
        not any(part.startswith('_') for part in path.parts)
        and process.is_valid_path(
            path=path,
            list_regex=[re.compile(r'^.*\.html$')],
            exclude_globs=build_config.exclude),
        (Path(path) for path in iglob(str(dir_src / '**/*'), recursive=True))
    )

    gen_copy = filter(
        lambda path: process.is_valid_path(
            path=path,
            list_regex=list_copy_regex,
            exclude_globs=build_config.exclude),
        (Path(path) for path in iglob(str(dir_src / '**/*'), recursive=True))
    )

    gen_path = chain(gen_html, gen_copy)

    gen_file_process_info = (
        FileProcessInfo(path=path, dir_src=dir_src, dir_dest=dir_dest)
        for path in gen_path
    )

    # Process each file
    for file_process_info in gen_file_process_info:
        if file_process_info.path.suffix == '.html':
            logger.info(f"Processing HTML file: {file_process_info.path}")
            process.build_html(file_process_info)
        else:
            logger.info(f"Copying file: {file_process_info.path}")
            process.copy_file(file_process_info)

    logger.success("✨ Build complete")
